import argparse
import hashlib
import logging
import os
import re
import subprocess
import sys
import threading
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import coloredlogs
from certbot.compat import misc

try:
    POSIX_MODE = True
    import grp
    import pwd
except ImportError:
    POSIX_MODE = False

LOGGER = logging.getLogger(__name__)
coloredlogs.install(logger=LOGGER)


def execute(
    command: Union[List[str], str],
    check: bool = True,
    shell: bool = False,
    env: Dict[str, str] = None,
    lock: Optional[threading.Lock] = None,
):
    if not env:
        env = os.environ.copy()
    env = env.copy()
    env["PYTHONUNBUFFERED "] = "1"

    call = subprocess.check_call if check else subprocess.call

    LOGGER.info(
        f"Launching command: {subprocess.list2cmdline(command) if isinstance(command, list) else command}"
    )
    sys.stdout.write("----------\n")
    sys.stdout.flush()

    error = None
    try:
        if lock:
            with lock:
                call(command, shell=shell, env=env)
        else:
            call(command, shell=shell, env=env)
    except subprocess.CalledProcessError as e:
        error = e

    sys.stdout.write("----------\n")
    sys.stdout.flush()

    if error:
        raise error


def fix_permissions(certificate_permissions: Dict[str, Any], target_path: str):
    files_mode = certificate_permissions.get("files_mode", 0o640)
    dirs_mode = certificate_permissions.get("dirs_mode", 0o750)

    os.chmod(target_path, dirs_mode)

    uid = -1
    gid = -1

    user = certificate_permissions.get("user")
    group = certificate_permissions.get("group")

    if (user or group) and not POSIX_MODE:
        LOGGER.warning(
            "Setting user and group for certificates/keys is not supported on Windows."
        )
    elif POSIX_MODE:
        if isinstance(user, int):
            uid = user
        elif isinstance(user, str):
            uid = pwd.getpwnam(user)[2]

        if isinstance(group, int):
            gid = group
        elif isinstance(group, str):
            gid = grp.getgrnam(group)[2]

        os.chown(target_path, uid, gid)  # type: ignore

    for root, dirs, files in os.walk(target_path):
        for path in dirs:
            os.chmod(os.path.join(root, path), dirs_mode)
        for path in files:
            os.chmod(os.path.join(root, path), files_mode)
        if POSIX_MODE:
            for path in files + dirs:
                os.chown(os.path.join(root, path), uid, gid)  # type: ignore


def configure_certbot_workspace(
    dnsrobocert_config: Dict[str, Any], directory_path: str
):
    live_path = os.path.join(directory_path, "archive")
    archive_path = os.path.join(directory_path, "live")

    if not os.path.exists(live_path):
        os.makedirs(live_path)
    if not os.path.exists(archive_path):
        os.makedirs(archive_path)

    certificate_permissions = dnsrobocert_config.get("acme", {}).get(
        "certs_permissions", {}
    )
    fix_permissions(certificate_permissions, live_path)
    fix_permissions(certificate_permissions, archive_path)


def digest(path: str):
    if not os.path.exists(path):
        return None

    with open(path, "rb") as file_h:
        config_data = file_h.read()

    md5 = hashlib.md5()
    md5.update(config_data)
    return md5.digest()


def normalize_lineage(domain: str):
    return re.sub(r"^\*\.", "", domain)


def validate_snap_environment(args: argparse.Namespace):
    """
    Verify that the provided CLI flags are compatible with an execution of
    DNSroboCert inside snap. No-op if we are not in a snap.
    """
    if not os.environ.get("SNAP"):
        return

    errors = []

    if not _is_valid_path_for_snap(args.config):
        errors.append(f"Invalid --config value: {args.config}")

    if not _is_valid_path_for_snap(args.directory):
        errors.append(f"Invalid --directory value: {args.directory}")

    for error in errors:
        LOGGER.error(error)

    if errors:
        LOGGER.error(
            "Snap DNSroboCert can only use non-hidden files and directories "
            "located in the user HOME folder by default."
        )
        LOGGER.error(
            "DNSroboCert can be granted to use the /etc directory with this command: "
            "snap connect dnsrobocert:etc"
        )

        sys.exit(1)


def get_default_args() -> Dict[str, str]:
    """
    Retrieve the default values of CLI flags depending on the execution context.
    """
    if os.environ.get("SNAP"):
        user_home = os.environ.get("SNAP_REAL_HOME", str(Path.home()))
        return {
            "config": os.path.join(user_home, "dnsrobocert/dnsrobocert.yml"),
            "configDesc": os.path.join(user_home, "dnsrobocert/dnsrobocert.yml"),
            "directory": os.path.join(user_home, "dnsrobocert/letsencrypt"),
            "directoryDesc": os.path.join(user_home, "dnsrobocert/letsencrypt"),
        }

    return {
        "config": os.path.join(os.getcwd(), "dnsrobocert.yml"),
        "configDesc": "$(pwd)/dnsrobocert.yml",
        "directory": misc.get_default_folder("config"),
        "directoryDesc": misc.get_default_folder("config"),
    }


def _is_valid_path_for_snap(path: str):
    """
    Check that the given path:
    * is relative to home_path and it is not composed of hidden parts (folder/files starting with "."),
    * or is located inside /etc if the snap is authorized to use /etc (connected to system_file interface).
    """
    home_path = os.environ.get("SNAP_REAL_HOME", str(Path.home()))

    # Check if the snap is authorized to use the /etc directory
    try:
        os.listdir("/etc")
        system_files_interface = True
    except PermissionError:
        system_files_interface = False

    return (
        os.path.abspath(path).startswith(home_path)
        and not any(part for part in _split_path(path) if part.startswith("."))
    ) or (system_files_interface and os.path.abspath(path).startswith("/etc/"))


def _split_path(path: str) -> List[str]:
    """
    Split the given path into an array containing each part composing the path.
    For instance: /etc/to/my/config.yml => ["/", "etc", "to", "my", "config.yml"].
    """
    all_parts: List[str] = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            all_parts.insert(0, parts[0])
            break
        elif parts[1] == path:  # sentinel for relative paths
            all_parts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            all_parts.insert(0, parts[1])
    return all_parts



import logging
import threading
import time
from contextlib import contextmanager
from random import random

import coloredlogs
import schedule

from dnsrobocert.core import certbot

LOGGER = logging.getLogger(__name__)
coloredlogs.install(logger=LOGGER)


@contextmanager
def worker(config_path: str, directory_path: str, lock: threading.Lock):
    stop_thread = threading.Event()

    schedule.every().day.at("12:00").do(
        _renew_job,
        config_path=config_path,
        directory_path=directory_path,
        lock=lock,
        stop_thread=stop_thread,
    )
    schedule.every().day.at("00:00").do(
        _renew_job,
        config_path=config_path,
        directory_path=directory_path,
        lock=lock,
        stop_thread=stop_thread,
    )

    _launch_background_jobs(stop_thread)

    try:
        yield
    finally:
        stop_thread.set()


def _launch_background_jobs(stop_thread: threading.Event, interval: int = 1):
    class ScheduleThread(threading.Thread):
        @classmethod
        def run(cls):
            while not stop_thread.is_set():
                schedule.run_pending()
                time.sleep(interval)

    continuous_thread = ScheduleThread()
    continuous_thread.start()


def _renew_job(
    config_path: str,
    directory_path: str,
    lock: threading.Lock,
    stop_thread: threading.Event,
):
    random_delay_seconds = 21600  # Random delay up to 12 hours
    wait_time = int(random() * random_delay_seconds)

    LOGGER.info("Automated execution: renew certificates if needed.")
    LOGGER.info(f"Random wait for this execution: {wait_time} seconds")

    interrupted = stop_thread.wait(wait_time)

    if not interrupted:
        certbot.renew(config_path, directory_path, lock)



import logging
import os
import re
from typing import Any, Dict, Optional, Set

import coloredlogs
import jsonschema
import pkg_resources
import yaml

from dnsrobocert.core import utils

LOGGER = logging.getLogger(__name__)
coloredlogs.install(logger=LOGGER)


def load(config_path: str) -> Optional[Dict[str, Any]]:
    if not os.path.exists(config_path):
        LOGGER.error(f"Configuration file {config_path} does not exist.")
        return None

    with open(config_path) as file_h:
        raw_config = file_h.read()

    raw_config = _inject_env_variables(raw_config)

    try:
        config = yaml.load(raw_config, Loader=yaml.SafeLoader)
    except BaseException:
        message = """
Error while validating dnsrobocert configuration:
Configuration file is not a valid YAML file.\
"""
        LOGGER.error(message)
        return None

    schema_path = pkg_resources.resource_filename("dnsrobocert", "schema.yml")
    with open(schema_path) as file_h:
        schema = yaml.load(file_h.read(), yaml.SafeLoader)

    if not config:
        message = """
Error while validating dnsrobocert configuration:
Configuration file is empty.\
"""
        LOGGER.error(message)
        return None

    try:
        jsonschema.validate(instance=config, schema=schema)
    except jsonschema.ValidationError as e:
        node = "/" + "/".join([str(item) for item in e.path])
        message = f"""\
Error while validating dnsrobocert configuration for node path {node}:
{e.message}.
-----
{raw_config}\
"""
        LOGGER.error(message)
        return None

    try:
        _values_conversion(config)
        _business_check(config)
    except ValueError as e:
        message = f"""\
Error while validating dnsrobocert configuration:
{str(e)}
-----
{raw_config}\
"""
        LOGGER.error(message)
        return None

    return config


def get_profile(config: Dict[str, Any], profile_name: str) -> Dict[str, Any]:
    profiles = [
        profile
        for profile in config.get("profiles", {})
        if profile["name"] == profile_name
    ]
    return profiles[0] if profiles else None


def get_certificate(config: Dict[str, Any], lineage: str) -> Optional[Dict[str, Any]]:
    certificates = [
        certificate
        for certificate in config.get("certificates", [])
        if get_lineage(certificate) == lineage
    ]
    return certificates[0] if certificates else None


def get_lineage(certificate_config: Dict[str, Any]) -> str:
    lineage = (
        certificate_config.get("name")
        if certificate_config.get("name")
        else utils.normalize_lineage(certificate_config.get("domains", [None])[0])
    )
    if not lineage:
        raise ValueError(
            f"Could not find the certificate name for certificate config: {certificate_config}"
        )

    return lineage


def get_acme_url(config: Dict[str, Any]) -> str:
    acme = config.get("acme", {})

    directory_url = acme.get("directory_url")
    if directory_url:
        return directory_url

    api_version = acme.get("api_version", 2)
    staging = acme.get("staging", False)

    if api_version < 2:
        domain = "acme-staging" if staging else "acme-v01"
    else:
        domain = "acme-staging-v02" if staging else "acme-v02"

    return f"https://{domain}.api.letsencrypt.org/directory"


def find_profile_for_lineage(config: Dict[str, Any], lineage: str) -> Dict[str, Any]:
    certificate = get_certificate(config, lineage)
    if not certificate:
        raise RuntimeError(
            f"Error, certificate named `{lineage}` could not be found in configuration."
        )
    profile_name = certificate.get("profile")
    if not profile_name:
        raise RuntimeError(
            f"Error, profile named `{lineage}` could not be found in configuration."
        )

    return get_profile(config, profile_name)


def _inject_env_variables(raw_config: str):
    def replace(match):
        entry = match.group(0)

        if "$${" in entry:
            return entry.replace("$${", "${")

        variable_name = match.group(1)
        if variable_name not in os.environ:
            raise ValueError(
                f"Error while parsing config: environment variable {variable_name} does not exist."
            )

        return os.environ[variable_name]

    return re.sub(r"\${1,2}{(\S+)}", replace, raw_config)


def _values_conversion(config: Dict[str, Any]):
    certs_permissions = config.get("acme", {}).get("certs_permissions", {})
    files_mode = certs_permissions.get("files_mode")
    if isinstance(files_mode, str):
        certs_permissions["files_mode"] = int(files_mode, 8)
    dirs_mode = certs_permissions.get("dirs_mode")
    if isinstance(dirs_mode, str):
        certs_permissions["dirs_mode"] = int(dirs_mode, 8)


def _business_check(config: Dict[str, Any]):
    profiles = [profile["name"] for profile in config.get("profiles", [])]
    lineages: Set[str] = set()
    for certificate_config in config.get("certificates", []):
        # Check that every certificate is associated to an existing profile
        profile = certificate_config.get("profile")
        lineage = get_lineage(certificate_config)
        if lineage:
            if profile not in profiles:
                raise ValueError(
                    f"Profile `{profile}` used by certificate `{lineage}` does not exist."
                )

            if lineage in lineages:
                raise ValueError(f"Certificate with name `{lineage}` is duplicated.")
            lineages.add(lineage)

    # Check that each files_mode and dirs_mode is a valid POSIX mode
    files_mode = config.get("acme", {}).get("certs_permissions", {}).get("files_mode")
    if files_mode and files_mode > 511:
        raise ValueError("Invalid files_mode `{0}` provided.".format(oct(files_mode)))
    dirs_mode = config.get("acme", {}).get("certs_permissions", {}).get("dirs_mode")
    if dirs_mode and dirs_mode > 511:
        raise ValueError("Invalid dirs_mode `{0}` provided.".format(oct(files_mode)))



import pkg_resources


def get_version():
    try:
        distribution = pkg_resources.get_distribution(__name__)
    except pkg_resources.DistributionNotFound:
        return "dev"
    else:
        return distribution.version


__version__ = get_version()



from typing import Any, Dict, Optional

import tldextract
from dns.exception import Timeout
from dns.rdatatype import RdataType
from dns.resolver import NXDOMAIN, NoAnswer, get_default_resolver
from lexicon.client import Client
from lexicon.config import ConfigResolver


def txt_challenge(
    certificate: Dict[str, Any],
    profile: Dict[str, Any],
    token: str,
    domain: str,
    action: str = "create",
):
    profile_name = profile["name"]
    provider_name = profile["provider"]
    provider_options = profile.get("provider_options", {})

    if not provider_options:
        print(
            f"No provider_options are defined for profile {profile_name}, "
            "any call to the provider API is likely to fail."
        )

    challenge_name = f"_acme-challenge.{domain}."
    if certificate.get("follow_cnames"):
        print(f"Trying to resolve the canonical challenge name for {challenge_name}")
        canonical_challenge_name = resolve_canonical_challenge_name(challenge_name)
        print(
            f"Canonical challenge name found for {challenge_name}: {canonical_challenge_name}"
        )
        challenge_name = canonical_challenge_name

        extracted = tldextract.extract(challenge_name)
        domain = ".".join([extracted.domain, extracted.suffix])

    config_dict = {
        "action": action,
        "domain": domain,
        "type": "TXT",
        "name": challenge_name,
        "content": token,
        "delegated": profile.get("delegated_subdomain"),
        "provider_name": provider_name,
        provider_name: provider_options,
    }

    ttl = profile.get("ttl")
    if ttl:
        config_dict["ttl"] = ttl

    lexicon_config = ConfigResolver()
    lexicon_config.with_dict(config_dict)

    Client(lexicon_config).execute()


def check_one_challenge(challenge: str, token: Optional[str]) -> bool:
    resolver = get_default_resolver()

    try:
        answers = resolver.query(challenge, "TXT")
    except (NXDOMAIN, NoAnswer):
        print(f"TXT {challenge} does not exist.")
        return False
    except Timeout as e:
        print(f"Timeout while trying to check TXT {challenge}: {e}")
        return False
    else:
        print(f"TXT {challenge} exists.")

    if token:
        validation_answers = [
            rdata
            for rdata in answers
            for txt_string in rdata.strings
            if txt_string.decode("utf-8") == token
        ]

        if not validation_answers:
            print(f"TXT {challenge} does not have the expected token value.")
            return False

        print(f"TXT {challenge} has the expected token value.")

    return True


def resolve_canonical_challenge_name(name: str) -> str:
    resolver = get_default_resolver()
    current_name = name
    visited = [current_name]

    while True:
        try:
            answer = resolver.resolve(current_name, rdtype=RdataType.CNAME)
            current_name = str(answer[0].target)
            if current_name in visited:
                resolution_map = " -> ".join([*visited, current_name])
                raise ValueError(
                    f"Error, CNAME resolution for {current_name} ended in an infinite loop!\n"
                    f"{resolution_map}"
                )
            visited.append(current_name)
        except (NXDOMAIN, NoAnswer):
            # No more CNAME in the chain, we have the final canonical_name
            return current_name
